from pymatex.listener import MatexASTVisitor
from pymatex.node import *


class IndexSearchVisitor(MatexASTVisitor.MatexASTVisitor):

    def __init__(self, data: dict):
        self.data = data
        self.results = {}
        self.seen_constants = {}
        self.seen_variables = {}
        self.bound_variables = set()

    def get_results(self):
        return self.results

    def visit_addition(self, addition_node: Addition):
        depth_lhs = addition_node.lhs.accept(self)
        depth_rhs = addition_node.rhs.accept(self)

        node_depth = max(depth_lhs, depth_rhs) + 1
        self.search(node_depth, NodeType.ADDITION)
        return node_depth

    def visit_constant(self, constant_node: Constant):
        node_depth = 0
        self.search_constant(node_depth, NodeType.CONSTANT, constant_node.value)
        return node_depth

    def visit_division(self, division_node: Division):
        depth_lhs = division_node.lhs.accept(self)
        depth_rhs = division_node.rhs.accept(self)

        node_depth = max(depth_lhs, depth_rhs) + 1
        self.search(node_depth, NodeType.DIVISION)
        return node_depth

    def visit_exponentiation(self, exponentiation_node: Exponentiation):
        depth_expr = exponentiation_node.lhs.accept(self)
        depth_exponent = exponentiation_node.rhs.accept(self)

        node_depth = max(depth_expr, depth_exponent) + 1
        self.search(node_depth, NodeType.EXPONENTIATION)
        return node_depth

    def visit_function(self, function_node: Function):
        depth = function_node.expression.accept(self)

        node_depth = depth + 1
        self.search(node_depth, NodeType.FUNCTION)
        return node_depth

    def visit_indexed_variable(self, indexed_variable_node: IndexedVariable):
        depth = indexed_variable_node.index.accept(self)

        node_depth = depth + 1
        self.search_variable(node_depth, NodeType.INDEXEDVARIABLE, indexed_variable_node.variable)
        return node_depth

    def visit_integral(self, integral_node: Integral):
        integral_node.variable.accept(self)
        integral_node.start_range.accept(self)
        integral_node.end_range.accept(self)
        self.add_bound_variable(integral_node.variable)
        depth_expression = integral_node.expression.accept(self)
        self.remove_bound_variable(integral_node.variable)

        node_depth = depth_expression + 1
        self.search(node_depth, NodeType.SUMMATION)
        return node_depth

    def visit_multiplication(self, multiplication_node: Multiplication):
        depth_lhs = multiplication_node.lhs.accept(self)
        depth_rhs = multiplication_node.rhs.accept(self)

        node_depth = max(depth_lhs, depth_rhs) + 1
        self.search(node_depth, NodeType.MULTIPLICATION)
        return node_depth

    def visit_negate(self, negate_node: Negate):
        depth = negate_node.node.accept(self)

        self.search(depth + 1, NodeType.NEGATE)
        return depth

    def visit_product(self, product_node: Product):
        product_node.variable.accept(self)
        product_node.start_range.accept(self)
        product_node.end_range.accept(self)
        self.add_bound_variable(product_node.variable)
        depth_expression = product_node.expression.accept(self)
        self.remove_bound_variable(product_node.variable)

        node_depth = depth_expression + 1
        self.search(node_depth, NodeType.SUMMATION)
        return node_depth

    def visit_subtraction(self, subtraction_node: Subtraction):
        depth_lhs = subtraction_node.lhs.accept(self)
        depth_rhs = subtraction_node.rhs.accept(self)

        node_depth = max(depth_lhs, depth_rhs) + 1
        self.search(node_depth, NodeType.SUBTRACTION)
        return node_depth

    def visit_summation(self, summation_node: Summation):
        summation_node.variable.accept(self)
        summation_node.start_range.accept(self)
        summation_node.end_range.accept(self)
        self.add_bound_variable(summation_node.variable)
        depth_expression = summation_node.expression.accept(self)
        self.remove_bound_variable(summation_node.variable)

        node_depth = depth_expression + 1
        self.search(node_depth, NodeType.SUMMATION)
        return node_depth

    def visit_variable(self, variable_node: Variable):
        node_depth = 0
        if str(variable_node.variable) in self.bound_variables:
            self.search_bound_variable(node_depth, NodeType.BOUNDVARIABLE, variable_node.variable)
        else:
            self.search_free_variable(node_depth, NodeType.VARIABLE, variable_node.variable)
        return node_depth

    def search(self, node_depth: int, node_type: NodeType):
        nodes = self.data.get(node_depth, None)
        if nodes is None:
            return

        objects = nodes.get(node_type, None)
        if objects:
            self.__add(objects, 100)

    def search_constant(self, node_depth: int, node_type: NodeType, external_data: str):
        nodes = self.data.get(node_depth, None)
        if nodes is None:
            return

        objects = nodes.get(node_type, None)
        if objects is None:
            return

        associated = objects.get(external_data, None)
        if associated:
            self.__add(associated, 100)
        else:
            for mathematical_objects in objects.values():
                self.__add(mathematical_objects, 70)

    def search_bound_variable(self, node_depth: int, node_type: NodeType, external_data: str):
        nodes = self.data.get(node_depth, None)
        if nodes is None:
            return

        objects = nodes.get(node_type, None)
        if objects is None:
            return

        associated = objects.get(external_data, None)
        if associated:
            self.__add(associated, 100)
        else:
            for mathematical_objects in objects.values():
                self.__add(mathematical_objects, 70)

        free_variables = nodes.get(node_type, None)
        if objects is None:
            return
        for mathematical_objects in free_variables.values():
            self.__add(mathematical_objects, 30)

    def search_free_variable(self, node_depth: int, node_type: NodeType, external_data: str):
        nodes = self.data.get(node_depth, None)
        if nodes is None:
            return

        objects = nodes.get(node_type, None)
        if objects is None:
            return

        associated = objects.get(external_data, None)
        if associated:
            self.__add(associated, 100)
        else:
            for mathematical_objects in objects.values():
                self.__add(mathematical_objects, 70)

    def __add(self, items, value):
        for item in items:
            new_value = self.results.get(item, 0) + value
            self.results[item] = new_value

    def add_bound_variable(self, variable: Variable):
        self.bound_variables.add(str(variable))

    def remove_bound_variable(self, variable: Variable):
        self.bound_variables.remove(str(variable))



from tests import BaseTest

from pymatex.node import *


class InequalityTests(BaseTest.BaseTest):

    def test_read_simple_inequality(self):
        ast = self.parse(r'\sum_{i < j} i')
        i = Variable('i')
        self.assertEqual(ast, InequalitySummation(Inequality(i, Variable('j'), '<'), None, i))

    def test_read_simple_inequality_with_constant(self):
        ast = self.parse(r'\sum_{0 < j} i')
        i = Variable('i')
        self.assertEqual(ast, InequalitySummation(Inequality(Constant('0'), Variable('j'), '<'), None, i))

    def test_read_multi_inequality(self):
        ast = self.parse(r'\sum_{0 < i \leq j} i')
        i = Variable('i')
        self.assertEqual(ast, InequalitySummation(Inequality(Inequality(Constant('0'), i, '<'), Variable('j'), '\\leq'), None, i))



from tests import BaseTest

from pymatex.node import Multiplication, Negate, Variable


class VariableTests(BaseTest.BaseTest):

    def test_read_variable(self):
        ast = self.parse('xyz')
        self.assertEqual(ast, Multiplication(Multiplication(Variable('x'), Variable('y')), Variable('z')))

    def test_write_variable(self):
        ast = self.parse('xyz')
        self.assertEqual(str(ast), '((x * y) * z)')

    def test_read_negative_variable(self):
        ast = self.parse('-x')
        self.assertEqual(ast, Negate(Variable('x')))

    def test_write_negative_variable(self):
        ast = self.parse('-x')
        self.assertEqual(str(ast), '-(x)')

    def test_read_special_variable(self):
        ast = self.parse('\pi')
        self.assertEqual(ast, Variable('\pi'))

    def test_write_special_variable(self):
        ast = self.parse('\pi')
        self.assertEqual(str(ast), '\pi')




from tests.node.Additive import *
from tests.node.Constant import *
from tests.node.Exponentiation import *
from tests.node.Inequality import *
from tests.node.Invalid import *
from tests.node.IterativeFunction import *
from tests.node.Function import *
from tests.node.Multiplicative import *
from tests.node.TaylorMaclaurin import *
from tests.node.Variable import *

from tests.node.Other import *



from pymatex.node import *


class MatexASTVisitor:
    def visit_addition(self, addition_node: Addition):
        addition_node.lhs.accept(self)
        addition_node.rhs.accept(self)
    
    def visit_constant(self, constant_node: Constant):
        pass
    
    def visit_division(self, division_node: Division):
        division_node.lhs.accept(self)
        division_node.rhs.accept(self)

    def visit_exponentiation(self, exponentiation_node: Exponentiation):
        exponentiation_node.lhs.accept(self)
        exponentiation_node.rhs.accept(self)
    
    def visit_function(self, function_node: Function):
        for argument in function_node.arguments:
            argument.accept(self)
    
    def visit_indexed_variable(self, indexed_variable_node: IndexedVariable):
        indexed_variable_node.index.accept(self)
    
    def visit_integral(self, integral_node: Integral):
        integral_node.variable.accept(self)
        integral_node.start_range.accept(self)
        integral_node.end_range.accept(self)
        integral_node.expression.accept(self)

    def visit_multiplication(self, multiplication_node: Multiplication):
        multiplication_node.lhs.accept(self)
        multiplication_node.rhs.accept(self)

    def visit_negate(self, negate_node: Negate):
        negate_node.node.accept(self)

    def visit_product(self, product_node: Product):
        if product_node.variable:
            product_node.variable.accept(self)
        product_node.start_range.accept(self)
        if product_node.end_range:
            product_node.end_range.accept(self)
        product_node.expression.accept(self)

    def visit_subtraction(self, subtraction_node: Subtraction):
        subtraction_node.lhs.accept(self)
        subtraction_node.rhs.accept(self)

    def visit_summation(self, summation_node: Summation):
        if summation_node.variable:
            summation_node.variable.accept(self)
        summation_node.start_range.accept(self)
        if summation_node.end_range:
            summation_node.end_range.accept(self)
        summation_node.expression.accept(self)

    def visit_variable(self, variable_node: Variable):
        pass



from pymatex.listener import MatexASTVisitor
from enum import Enum, auto


class NodeType(Enum):
    ADDITION = auto()
    BOUNDVARIABLE = auto()
    CONSTANT = auto()
    DIVISION = auto()
    EXPONENTIATION = auto()
    FUNCTION = auto()
    INDEXEDVARIABLE = auto()
    INTEGRAL = auto()
    MULTIPLICATION = auto()
    NEGATE = auto()
    PRODUCT = auto()
    SUBTRACTION = auto()
    SUMMATION = auto()
    VARIABLE = auto()


class Node:
    def accept(self, visitor: MatexASTVisitor):
        pass



from pymatex.listener import MatexASTVisitor
from pymatex.node.Variable import Variable


class IndexedVariable(Variable):

    def __init__(self, variable: str, index):
        super().__init__(variable)

        self.index = index

    def __eq__(self, other):
        return super().__eq__(other) and self.index == other.index

    def __str__(self):
        return '{}_{{{}}}'.format(self.variable, self.index)

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_indexed_variable(self)



from tests import BaseTest

from pymatex.node import Addition, Constant, Func, Function, Multiplication, Variable


class FunctionTests(BaseTest.BaseTest):

    def test_read_sqrt(self):
        ast = self.parse('\\sqrt{2x + 3}')
        self.assertEqual(ast, Function(Func.SQRT, Addition(Multiplication(Constant('2'), Variable('x')), Constant('3'))))

    def test_write_sqrt(self):
        ast = self.parse('\\sqrt{2x + 3}')
        self.assertEqual(str(ast), '\\sqrt{((2 * x) + 3)}')

    def test_read_sin(self):
        ast = self.parse('\\sin{2x + 3}')
        self.assertEqual(ast, Function(Func.SIN, Addition(Multiplication(Constant('2'), Variable('x')), Constant('3'))))

    def test_read_alternative_sin(self):
        ast = self.parse('sin(2x + 3)')
        self.assertEqual(ast, Function(Func.SIN, Addition(Multiplication(Constant('2'), Variable('x')), Constant('3'))))

    def test_write_sin(self):
        ast = self.parse('\\sin{2x + 3}')
        self.assertEqual(str(ast), '\\sin{((2 * x) + 3)}')

    def test_read_factorial(self):
        ast = self.parse('x!')
        self.assertEqual(ast, Function(Func.FACTORIAL, Variable('x')))

        ast = self.parse('3!')
        self.assertEqual(ast, Function(Func.FACTORIAL, Constant('3')))

        ast = self.parse('(2 * x)!')
        self.assertEqual(ast, Function(Func.FACTORIAL, Multiplication(Constant('2'), Variable('x'))))

        ast = self.parse('4 * x!')
        self.assertEqual(ast, Multiplication(Constant('4'), Function(Func.FACTORIAL, Variable('x'))))

    def test_write_factorial(self):
        ast = self.parse('x!')
        self.assertEqual(str(ast), '(x)!')

    def test_read_absolute(self):
        ast = self.parse('|x + 3|')
        self.assertEqual(ast, Function(Func.ABS, Addition(Variable('x'), Constant('3'))))

    def test_write_absolute(self):
        ast = self.parse('|x + 3|')
        self.assertEqual(str(ast), '|(x + 3)|')

    def test_read_integer_division(self):
        ast = self.parse('n|m')
        self.assertEqual(ast, Function(Func.EXACTDIVISION, Variable('n'), Variable('m')))

    def test_read_n_choose_k(self):
        ast = self.parse(r'\binom{x}{y}')
        self.assertEqual(ast, Function(Func.BINOMIAL, Variable('x'), Variable('y')))

    def test_read_zeta(self):
        ast = self.parse(r'\zeta(x)')
        self.assertEqual(ast, Function(Func.ZETA, Variable('x')))

    def test_read_elliptic_curve(self):
        ast = self.parse('sn(u, m)')
        self.assertEqual(ast, Function(Func.ESIN, Variable('u'), Variable('m')))

        ast = self.parse('cn(u, m)')
        self.assertEqual(ast, Function(Func.ECOS, Variable('u'), Variable('m')))

        ast = self.parse('dn(u, m)')
        self.assertEqual(ast, Function(Func.EDELTAAMPLITUDE, Variable('u'), Variable('m')))

    def test_read_arc_elliptic_curve(self):
        ast = self.parse('arcsn(u, m)')
        self.assertEqual(ast, Function(Func.AESIN, Variable('u'), Variable('m')))

        ast = self.parse('arccn(u, m)')
        self.assertEqual(ast, Function(Func.AECOS, Variable('u'), Variable('m')))

        ast = self.parse('arcdn(u, m)')
        self.assertEqual(ast, Function(Func.AEDELTAAMPLITUDE, Variable('u'), Variable('m')))

    def test_read_neville_theta_function(self):
        ast = self.parse(r'\theta_{c}(z, m)')
        self.assertEqual(ast, Function(r'\theta_{c}', Variable('z'), Variable('m')))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='pymatex',
    version='0.0.10',
    author='Gawaboumga',
    description='Parser which allows to parse mathematical expressions written in LaTeX',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/Gawaboumga/PyMatex',
    packages=setuptools.find_packages(),
    install_requires=['antlr4-python3-runtime'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from tests import BaseTest

from pymatex.node import Addition, Constant, Negate, Subtraction, Variable


class AdditiveTests(BaseTest.BaseTest):

    def test_read_addition_of_constants(self):
        ast = self.parse('3 + 2')
        self.assertEqual(ast, Addition(Constant('3'), Constant('2')))

    def test_write_addition_of_constants(self):
        ast = self.parse('3 + 2')
        self.assertEqual(str(ast), '(3 + 2)')

    def test_read_addition_of_multiple_constants(self):
        ast = self.parse('3 + 2 + 5')
        self.assertEqual(ast, Addition(Addition(Constant('3'), Constant('2')), Constant('5')))

        ast = self.parse('3+2+5')
        self.assertEqual(ast, Addition(Addition(Constant('3'), Constant('2')), Constant('5')))

    def test_read_addition_of_variables(self):
        ast = self.parse('x + y')
        self.assertEqual(ast, Addition(Variable('x'), Variable('y')))

        ast = self.parse('x+y')
        self.assertEqual(ast, Addition(Variable('x'), Variable('y')))

    def test_write_addition_of_variables(self):
        ast = self.parse('x+y')
        self.assertEqual(str(ast), '(x + y)')

    def test_read_addition_of_constant_and_variable(self):
        ast = self.parse('n+1')
        self.assertEqual(ast, Addition(Variable('n'), Constant('1')))

    def test_read_subtraction_of_constants(self):
        ast = self.parse('3 - 2')
        self.assertEqual(ast, Subtraction(Constant('3'), Constant('2')))

    def test_write_subtraction_of_constants(self):
        ast = self.parse('3 - 2')
        self.assertEqual(str(ast), '(3 - 2)')

    def test_read_subtraction_of_multiple_constants(self):
        ast = self.parse('3 - 2 - 5')
        self.assertEqual(ast, Subtraction(Subtraction(Constant('3'), Constant('2')), Constant('5')))

    def test_read_subtraction_of_variables(self):
        ast = self.parse('x - y')
        self.assertEqual(ast, Subtraction(Variable('x'), Variable('y')))

    def test_write_subtraction_of_variables(self):
        ast = self.parse('x-y')
        self.assertEqual(str(ast), '(x - y)')

    def test_read_subtraction_of_negate_variable(self):
        ast = self.parse('x - -y')
        self.assertEqual(ast, Subtraction(Variable('x'), Negate(Variable('y'))))

    def test_read_substraction_of_constant_and_variable(self):
        ast = self.parse('n-1')
        self.assertEqual(ast, Subtraction(Variable('n'), Constant('1')))




from pymatex.node.Addition import *
from pymatex.node.Constant import *
from pymatex.node.Division import *
from pymatex.node.Exponentiation import *
from pymatex.node.Function import *
from pymatex.node.IndexedVariable import *
from pymatex.node.Inequality import *
from pymatex.node.Integral import *
from pymatex.node.Multiplication import *
from pymatex.node.Negate import *
from pymatex.node.Node import *
from pymatex.node.Product import *
from pymatex.node.Subtraction import *
from pymatex.node.Summation import *
from pymatex.node.Variable import *



from tests import BaseTest

from pymatex.node import *


class IterativeFunctionTests(BaseTest.BaseTest):

    def test_read_simple_summation(self):
        ast = self.parse(r'\sum_{i=0}^{\infty} (i + 1)')
        self.assertEqual(ast, Summation(Variable('i'), Constant('0'), Constant('\\infty'), Addition(Variable('i'), Constant('1'))))

    def test_write_simple_summation(self):
        ast = self.parse(r'\sum_{i=0}^{\infty} (i + 1)')
        self.assertEqual(str(ast), '\sum_{i=0}^{\infty}{(i + 1)}')

    def test_read_nested_summation(self):
        ast = self.parse(r'\sum_{i = 1}^{\infty}\sum_{j = 1}^{\infty} (i + j)')
        self.assertEqual(ast, Summation(Variable('i'), Constant('1'), Constant('\\infty'),
                                        Summation(Variable('j'), Constant('1'), Constant('\\infty'),
                                                  Addition(Variable('i'), Variable('j')))))

    def test_read_simple_product(self):
        ast = self.parse(r'\prod_{i=0}^{\infty} (i + 1)')
        self.assertEqual(ast, Product(Variable('i'), Constant('0'), Constant('\\infty'), Addition(Variable('i'), Constant('1'))))

    def test_write_simple_product(self):
        ast = self.parse(r'\prod_{i=0}^{\infty} (i + 1)')
        self.assertEqual(str(ast), '\\prod_{i=0}^{\\infty}{(i + 1)}')

    def test_read_addition_of_summation_and_product(self):
        ast = self.parse(r'\sum_{i=0}^{\infty} i + \prod_{i=0}^{\infty} i')

        summation = Summation(Variable('i'), Constant('0'), Constant('\\infty'), Variable('i'))
        product = Product(Variable('i'), Constant('0'), Constant('\\infty'), Variable('i'))
        self.assertEqual(ast, Addition(summation, product))

    def test_read_subtraction_of_product_and_summation(self):
        ast = self.parse(r'\prod_{i=0}^{\infty} i - \sum_{i=0}^{\infty} i')

        summation = Summation(Variable('i'), Constant('0'), Constant('\\infty'), Variable('i'))
        product = Product(Variable('i'), Constant('0'), Constant('\\infty'), Variable('i'))
        self.assertEqual(ast, Subtraction(product, summation))

    def test_read_multiplication_of_summations(self):
        ast = self.parse(r'\sum_{i=0}^{\infty} i * \sum_{i=0}^{\infty} i * \sum_{i=0}^{\infty} i')

        summation = Summation(Variable('i'), Constant('0'), Constant('\\infty'), Variable('i'))
        self.assertEqual(ast, Multiplication(Multiplication(summation, summation), summation))

    def test_read_integral(self):
        ast = self.parse(r'\int_{a}^{b} 3x dx')

        summation = Integral(Variable('x'), Variable('a'), Variable('b'), Multiplication(Constant('3'), Variable('x')))
        self.assertEqual(ast, summation)

    def test_read_nested_integral(self):
        ast = self.parse(r'\int_{a}^{b} \int_{c}^{d} 3x y dy dx')

        internal = Integral(Variable('y'), Variable('c'), Variable('d'), Multiplication(Multiplication(Constant('3'), Variable('x')), Variable('y')))
        outer = Integral(Variable('x'), Variable('a'), Variable('b'), internal)
        self.assertEqual(ast, outer)

    def test_read_function_call_zeta(self):
        ast = self.parse(r'\sum_{n=0}^{\infty} \zeta(2n) x^{2n}')
        two_n = Multiplication(Constant('2'), Variable('n'))
        self.assertEqual(ast, Summation(Variable('n'), Constant('0'), Constant('\\infty'), Multiplication(Function(Func.ZETA, two_n), Exponentiation(Variable('x'), two_n))))




import unittest
from pymatex.search import SimilarityMatrix


class SimilarityMatrixTests(unittest.TestCase):

    def test_read(self):
        s = SimilarityMatrix(path='tests/search/resources/SimilarityMatrixTests.txt')




from antlr4.error.ErrorListener import ErrorListener


class ExceptionErrorListener(ErrorListener):

    def __init__(self):
        super(ExceptionErrorListener, self).__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception("{}: syntax error {}".format(column, msg))

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        pass

    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        pass

    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        pass
